'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var Cluster = _interopDefault(require('cluster'));
var Mongoose = _interopDefault(require('mongoose'));
var Mexp = _interopDefault(require('mongoose-elasticsearch-xp'));
var graphqlCompose = require('graphql-compose');
var Logger = _interopDefault(require('@foundationjs/logger'));
require('elasticsearch');
var node8 = require('graphql-compose-mongoose/node8');

function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) {
  try {
    var info = gen[key](arg);
    var value = info.value;
  } catch (error) {
    reject(error);
    return;
  }

  if (info.done) {
    resolve(value);
  } else {
    Promise.resolve(value).then(_next, _throw);
  }
}

function _asyncToGenerator(fn) {
  return function () {
    var self = this,
        args = arguments;
    return new Promise(function (resolve, reject) {
      var gen = fn.apply(self, args);

      function _next(value) {
        asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value);
      }

      function _throw(err) {
        asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err);
      }

      _next(undefined);
    });
  };
}

function _slicedToArray(arr, i) {
  return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _nonIterableRest();
}

function _arrayWithHoles(arr) {
  if (Array.isArray(arr)) return arr;
}

function _iterableToArrayLimit(arr, i) {
  var _arr = [];
  var _n = true;
  var _d = false;
  var _e = undefined;

  try {
    for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _nonIterableRest() {
  throw new TypeError("Invalid attempt to destructure non-iterable instance");
}

function capitalize(name) {
  if (!name) {
    return '';
  }

  return name[0].toUpperCase() + name.slice(1);
}
function getComposerName(model, modelName) {
  const composer = model.composer;
  const name = modelName || model.name;

  if (composer) {
    return capitalize(composer);
  }

  const singular = (name[name.length - 1] === 's' ? name.slice(0, -1) : name).replace(/-([a-z])/g, (s, v) => v.toUpperCase());
  return capitalize(singular);
}

const makeComposers = ({
  schemaComposer,
  models: Models,
  scopeAccessResolver,
  exclude,
  refMap
}) => {
  const composers = {};
  Object.keys(Models).map(name => {
    const model = Models[name];
    const composerName = getComposerName(model, name);
    const composer = node8.composeWithMongoose(model, {
      name: composerName
    });
    composer.removeField('password');
    composers[composerName] = composer;
    composers[composerName].model = model;
    return {
      model,
      composer
    };
  }).forEach(({
    model,
    composer
  }) => {
    const fields = Object.keys(model.schema.tree).map(key => [key, model.schema.tree[key]]);
    const refFields = [];

    const findNestedRefFields = (fieldsToTest, path) => {
      // Find all nested ref
      fieldsToTest.filter(field => {
        // Test the field, or the first element of the field if it's an array
        const toCheck = Array.isArray(field[1]) ? field[1][0] : field[1];

        if (field[0] === 'id') {
          // If the field name is "id" then it's a Mongoose internal field
          return false;
        }

        return !( // If it's a function, then it's a standard Mongoose type
        toCheck instanceof Function // If a "schemaName" is present, then it's a standard field
        || !!toCheck.schemaName // If "type" is a function and "ref" is not a string, then it's a Mongoose complex field
        || typeof toCheck.ref !== 'string' && toCheck.type instanceof Function);
      }).forEach(field => {
        const array = Array.isArray(field[1]);
        const toCheck = array ? field[1][0] : field[1];

        if (typeof toCheck.ref === 'string') {
          // If the ref argument is a string, then we have a ref field
          refFields.push({
            field: field[0],
            ref: toCheck.ref,
            path,
            array
          });
        } else {
          // Else, we have a nested object or array of nested object, so check all fields again
          findNestedRefFields(Object.entries(toCheck), [...path, field[0]]);
        }
      });
    };

    findNestedRefFields(fields, []); // Add all ref resolvers

    refFields.forEach(field => {
      let refComposer = composer; // Get the correct composer for deeply nested fields

      if (field.path.length) {
        field.path.forEach(path => {
          refComposer = refComposer.getFieldTC(path);
        });
      } // Get the related composer by the ref


      const relatedComposerName = getComposerName(refMap[field.ref]); // Add the relation

      refComposer.addRelation(field.field, {
        resolver: () => composers[relatedComposerName].getResolver(field.array ? 'findByIds' : 'findById'),
        prepareArgs: {
          [field.array ? '_ids' : '_id']: source => source[field.field]
        }
      });
    });
  });
  const resolvers = {
    Query: {
      get: ['findOne', ''],
      list: ['findMany', 'List'],
      count: ['count', 'Count']
    },
    Mutation: {
      create: ['createOne', 'Create'],
      createMany: ['createMany', 'CreateMany'],
      update: ['updateById', 'Update'],
      delete: ['removeById', 'Delete']
    }
  };
  Object.keys(composers).filter(name => !exclude || exclude.indexOf(name) < 0).map(name => [name, composers[name]]).forEach(([name, composer]) => {
    const lowercase = name[0].toLowerCase() + name.slice(1);

    const createResolver = type => resolver => {
      const _resolvers$type$resol = _slicedToArray(resolvers[type][resolver], 2),
            method = _resolvers$type$resol[0],
            suffix = _resolvers$type$resol[1];

      composer[resolver] = scopeAccessResolver(composers)(composer.getResolver(method), name, resolver); // eslint-disable-line

      composer[resolver].unscoped = composer.getResolver(method); // eslint-disable-line

      composer[resolver].args = composer.getResolver(method).getArgs(); // eslint-disable-line

      composer[resolver].type = composer.getResolver(method); // eslint-disable-line

      schemaComposer[type].addFields({
        [lowercase + suffix]: composer[resolver]
      });
    };

    Object.keys(resolvers).forEach(type => {
      Object.keys(resolvers[type]).forEach(createResolver(type));
    });
  });
  return composers;
};

Mongoose.Promise = global.Promise;

function composerName(model) {
  return getComposerName(model);
}

function modelName(name) {
  return name.replace(/^([a-z])/, (f, l) => l.toUpperCase());
}
/**
 * Return a list of pre-configured composers
 * @param {object} composers - Map of composers
 * @param {object} context - Request context
 * @returns {object} - Map of context-injected composers
 */


function scopeComposers(composers, context) {
  const scopedComposers = {};
  const queryResolvers = ['get', 'list', 'count'];
  const recordResolvers = ['create', 'update'];
  const recordsResolvers = ['createMany'];
  const idResolvers = ['delete'];
  Object.keys(composers).forEach(key => {
    const composer = composers[key];
    scopedComposers[key] = {
      model: composers[key].model
    };
    queryResolvers.forEach(rs => {
      scopedComposers[key][rs] = (filter, {
        skip = undefined,
        sort = undefined,
        limit = undefined
      } = {}) => composer[rs].resolve({
        args: {
          filter,
          skip,
          sort,
          limit
        },
        context
      });

      scopedComposers[key][rs].unscoped = (filter, {
        skip = undefined,
        sort = undefined,
        limit = undefined
      } = {}) => composer[rs].unscoped.resolve({
        args: {
          filter,
          skip,
          sort,
          limit
        },
        context
      });
    });
    recordResolvers.forEach(rs => {
      scopedComposers[key][rs] = record => composer[rs].resolve({
        args: {
          record
        },
        context
      });

      scopedComposers[key][rs].unscoped = record => composer[rs].unscoped.resolve({
        args: {
          record
        },
        context
      });
    });
    recordsResolvers.forEach(rs => {
      scopedComposers[key][rs] = records => composer[rs].resolve({
        args: {
          records
        },
        context
      });

      scopedComposers[key][rs].unscoped = records => composer[rs].unscoped.resolve({
        args: {
          records
        },
        context
      });
    });
    idResolvers.forEach(rs => {
      scopedComposers[key][rs] = _id => composer[rs].resolve({
        args: {
          _id
        },
        context
      });

      scopedComposers[key][rs].unscoped = _id => composer[rs].unscoped.resolve({
        args: {
          _id
        },
        context
      });
    });
  });
  return scopedComposers;
}
/**
 * Add a field to a composer, with scoped request
 * @param {string} field - Name of the field to add
 * @param {string} kind - Whether it's a field, a query or a mutation
 * @param {*} to - Composer to add the field to
 * @param {*} composers - Map of composers to inject
 */


function addField(field, kind, to, composers) {
  const config = kind[field];
  to.addFields({
    [field]: {
      projection: config.needs || config.projection,
      type: config.type || (config.extends ? config.extends.type : undefined),
      args: config.args || (config.extends ? config.extends.args : undefined),
      resolve: function () {
        var _resolve = _asyncToGenerator(function* (source, args, context, info) {
          return config.resolve({
            source,
            args,
            context,
            info,
            composers: scopeComposers(composers, context)
          });
        });

        function resolve(_x, _x2, _x3, _x4) {
          return _resolve.apply(this, arguments);
        }

        return resolve;
      }()
    }
  });
}

function makeEsExtend(fields) {
  const esFields = {};
  Object.keys(fields).forEach(f => {
    const field = fields[f];
    esFields[f] = {
      es_value: field.value || field.es_value,
      es_type: field.type || field.es_type
    };
  });
  return esFields;
}

const Types = Object.assign({}, Mongoose.Schema.Types, {
  Map: Mongoose.Schema.Types.Mixed
});
/**
 * Class Persistence: initialize a Persistence configuration for a Foundation App.
 */

class Persistence {
  /**
   * Constructor
   * @param {PersistenceConfiguration} config - Optional config object
   */
  constructor(config) {
    this.configureIO = ({
      io
    }) => {
      const logger = this.logger;
      logger.info('New Socket.IO service available. Connecting...');
      this.io = io;
      io.on('connection', socket => {
        socket.join('updates');
      });
      logger.info('Socket.IO connected.');
    };

    if (config) {
      this.init(config);
    }
  }
  /**
   * Function configuring the Persistence instance
   * @param {PersistenceConfiguration} config - Configuration
   * @returns {boolean} - Successful initialization
   */


  init(config) {
    var _this = this;

    return _asyncToGenerator(function* () {
      const modelsConfig = config.models,
            query = config.query,
            mutation = config.mutation,
            log = config.log,
            endpoint = config.endpoint,
            elasticsearch$$1 = config.elasticsearch; // Prepare the logger

      const logConfig = log || {};

      if (!Cluster.isMaster) {
        logConfig.filename = logConfig.filename && `[${Cluster.worker.id}]_${logConfig.filename}`;
      } // Prepare the logger


      _this.logger = new Logger('Persistence', logConfig);
      const logger = _this.logger;

      if (!Cluster.isMaster) {
        logger.info(`Cluster Mode! Instance ${Cluster.worker.id} running`);
        logger.level = 'error';
      }

      logger.info('Initializing Persistence');
      const scopes = {};
      const models = {};
      const refMap = {};

      if (!modelsConfig.length) {
        logger.error('Cannot initialize Persistence without any model!');
        return false;
      }

      logger.info(`Found ${modelsConfig.length} models. Importing...`);
      modelsConfig.forEach(model => {
        scopes[composerName(model)] = model.scope;
        refMap[model.name] = model;
        let params;

        if (model.elasticsearch && model.elasticsearch.fields) {
          params = {
            es_extend: makeEsExtend(model.elasticsearch.fields)
          };
        }

        const schema = new Mongoose.Schema(model.schema, params);
        const name = modelName(model.name);
        schema.post('save', doc => {
          logger.debug(`${name} updated`);

          if (model.onPostSave) {
            model.onPostSave(doc);
          }

          if (_this.io) {
            _this.io.to('updates').emit(`${name.toLowerCase()}-updated`, doc._id);

            _this.io.to('updates').emit(`${name.toLowerCase()}-saved`, doc._id);
          }
        });
        schema.post('remove', doc => {
          logger.debug(`${name} updated`);

          if (model.onPostRemove) {
            model.onPostRemove(doc);
          }

          if (_this.io) {
            _this.io.to('updates').emit(`${name.toLowerCase()}-updated`, doc._id);

            _this.io.to('updates').emit(`${name.toLowerCase()}-removed`, doc._id);
          }
        });

        if (model.elasticsearch) {
          // Get the list of path to populate
          const paths = [];

          if (model.elasticsearch.fields) {
            Object.values(model.elasticsearch.fields).forEach(field => {
              if (field.es_populate) {
                paths.push(...Object.entries(field.es_populate).filter(p => p[1]).map(([path]) => path));
              }
            });
          }

          schema.post('save',
          /*#__PURE__*/
          function () {
            var _ref = _asyncToGenerator(function* (doc, next) {
              if (paths.length) {
                let exec = doc;
                paths.forEach(path => {
                  exec = exec.populate(path);
                });
                yield doc.execPopulate();
              }

              next();
            });

            return function (_x5, _x6) {
              return _ref.apply(this, arguments);
            };
          }());
          const mexpParams = {
            hydrate: true
          };

          if (elasticsearch$$1) {
            if (elasticsearch$$1.host) {
              mexpParams.host = elasticsearch$$1.host;
            }

            if (elasticsearch$$1.auth) {
              mexpParams.auth = elasticsearch$$1.auth;
            }

            if (elasticsearch$$1.port) {
              mexpParams.port = elasticsearch$$1.port;
            }

            if (elasticsearch$$1.protocol) {
              mexpParams.protocol = elasticsearch$$1.protocol;
            }

            if (elasticsearch$$1.prefix) {
              mexpParams.index = `${elasticsearch$$1.prefix}_${name}`.toLowerCase();
              mexpParams.type = `${elasticsearch$$1.prefix}_${name}`.toLowerCase();
            }
          }

          schema.plugin(Mexp, mexpParams);
        }

        models[name] = Mongoose.model(model.name, schema, model.collection); // Force the collection name if provided

        models[name].composer = model.composer;
        logger.info(`Model '${name}' imported.`);
      });

      const scopeAccessResolver = composers => (resolver, name, type) => resolver.wrapResolve(next =>
      /*#__PURE__*/
      function () {
        var _ref2 = _asyncToGenerator(function* ({
          source,
          args,
          context,
          info
        }) {
          if (!scopes[name] || !(scopes[name] instanceof Function)) {
            return next({
              source,
              args,
              context,
              info
            });
          }

          const newArgs = (yield scopes[name]({
            source,
            args,
            context,
            info,
            models,
            composers: scopeComposers(composers, context),
            type
          })) || args;
          return next({
            source,
            args: newArgs,
            context,
            info
          });
        });

        return function (_x7) {
          return _ref2.apply(this, arguments);
        };
      }());

      const composers = makeComposers({
        schemaComposer: graphqlCompose.schemaComposer,
        models,
        scopeAccessResolver,
        refMap
      });

      if (query) {
        logger.info('Importing custom Query...');
        query({
          schemaComposer: graphqlCompose.schemaComposer,
          composers,
          scopeAccessResolver: scopeAccessResolver(composers),
          models
        });
        logger.info('Imported.');
      }

      if (mutation) {
        logger.info('Importing custom Mutation...');
        mutation({
          schemaComposer: graphqlCompose.schemaComposer,
          composers,
          scopeAccessResolver: scopeAccessResolver(composers),
          models
        });
        logger.info('Imported.');
      }

      const typeComposers = {};
      Object.keys(composers).forEach(comp => {
        typeComposers[comp] = composers[comp]; // Deprecated

        typeComposers[`${comp}TC`] = composers[comp];
      });
      _this.typeComposers = typeComposers;
      _this.Query = graphqlCompose.schemaComposer.Query;
      _this.Mutation = graphqlCompose.schemaComposer.Mutation;
      logger.info('Adding custom fields to models...');
      modelsConfig.forEach(model => {
        const fields = model.fields;

        if (fields) {
          const fieldsConfig = fields({
            schemaComposer: graphqlCompose.schemaComposer,
            composers,
            scopeAccessResolver: scopeAccessResolver(composers),
            models,
            typeComposers
          });

          if (fieldsConfig) {
            if (fieldsConfig.fields) {
              Object.keys(fieldsConfig.fields).forEach(field => addField(field, fieldsConfig.fields, composers[composerName(model)], composers));
            }

            if (fieldsConfig.queries) {
              Object.keys(fieldsConfig.queries).forEach(field => addField(field, fieldsConfig.queries, graphqlCompose.schemaComposer.Query, composers));
            }

            if (fieldsConfig.mutations) {
              Object.keys(fieldsConfig.mutations).forEach(field => addField(field, fieldsConfig.mutations, graphqlCompose.schemaComposer.Mutation, composers));
            }
          }
        }
      });
      logger.info('Done.');
      logger.info('Building GraphQL schema...');
      _this.schema = graphqlCompose.schemaComposer.buildSchema();
      logger.info('GraphQL schema available.');

      if (endpoint) {
        return _this.connect({
          endpoint
        });
      }

      return true;
    })();
  }

  connect({
    endpoint
  }) {
    var _this2 = this;

    return _asyncToGenerator(function* () {
      return new Promise((resolve, reject) => {
        const logger = _this2.logger;
        logger.info(`Connecting to Mongo at ${endpoint}.`);
        Mongoose.connect(endpoint, {
          useNewUrlParser: true
        }, () => {
          logger.info('Mongo connected.');
          resolve(true);
        }, error => {
          reject(error);
        });
      });
    })();
  }

}

exports.Types = Types;
exports.default = Persistence;
